import struct,random,base64,hashlib,typing,select,json
from typing import Union
from .session import Session
from . import ModuleWrapper
from http import HTTPStatus
from datetime import timedelta

def WebsocketFrameProperty(func):    
    @property
    def wrapper(self):
        return getattr(self,'_' + func.__name__)
    @wrapper.setter
    def wrapper(self,value):
        return setattr(self,'_' + func.__name__,value)
    return wrapper

class WebsocketConnectionClosedException(Exception):
    def __init__(self, is_requested: bool):
        self.is_requested = is_requested
        super().__init__(
            f'Client {"requested" if is_requested else "UNEXPECTLY"} closed connection')

class WebsocketFrame(object):
    '''Provides docstrings and converters for Websocket Frame bits'''
    @staticmethod
    def bytes(d) -> bytearray:
        '''Converts any datatype to a mutable bytearray'''
        if isinstance(d, str):
            return d.encode()
        if isinstance(d, bytearray) or isinstance(d, bytes):
            return d
        return WebsocketFrame.bytes(json.dumps(d))

    def __init__(self, FIN=1, RSV1=0, RSV2=0, RSV3=0, OPCODE=1, MASK=0, PAYLOAD_LENGTH=0, MASKEY=0, PAYLOAD=b''):
        self._FIN, self._RSV1, self._RSV2, self._RSV3, self._OPCODE, self._MASK, self._PAYLOAD_LENGTH, self._MASKEY, self._PAYLOAD = FIN, RSV1, RSV2, RSV3, OPCODE, MASK, PAYLOAD_LENGTH, MASKEY, self.bytes(
            PAYLOAD)
        if not self._PAYLOAD_LENGTH:
            self._PAYLOAD_LENGTH = len(self._PAYLOAD)
        # Re-applie payload length if not given
        super().__init__()

    @WebsocketFrameProperty
    def FIN(self):
        '''     FIN:  1 bit

      Indicates that this is the final fragment in a message.  The first
      fragment MAY also be the final fragment.'''
        pass

    @WebsocketFrameProperty
    def RSV1(self):
        '''   RSV1, RSV2, RSV3:  1 bit each

      MUST be 0 unless an extension is negotiated that defines meanings
      for non-zero values.  If a nonzero value is received and none of
      the negotiated extensions defines the meaning of such a nonzero
      value, the receiving endpoint MUST _Fail the WebSocket
      Connection_.'''
        pass

    @WebsocketFrameProperty
    def RSV2(self):
        '''   RSV1, RSV2, RSV3:  1 bit each

      MUST be 0 unless an extension is negotiated that defines meanings
      for non-zero values.  If a nonzero value is received and none of
      the negotiated extensions defines the meaning of such a nonzero
      value, the receiving endpoint MUST _Fail the WebSocket
      Connection_.'''
        pass

    @WebsocketFrameProperty
    def RSV3(self):
        '''   RSV1, RSV2, RSV3:  1 bit each

      MUST be 0 unless an extension is negotiated that defines meanings
      for non-zero values.  If a nonzero value is received and none of
      the negotiated extensions defines the meaning of such a nonzero
      value, the receiving endpoint MUST _Fail the WebSocket
      Connection_.'''
        pass

    @WebsocketFrameProperty
    def OPCODE(self):
        '''Opcode:  4 bits

      Defines the interpretation of the "Payload data".  If an unknown
      opcode is received, the receiving endpoint MUST _Fail the
      WebSocket Connection_.  The following values are defined.You can use those stored
      in `Websocket` class as well

      -  %x0 denotes a continuation frame

      -  %x1 denotes a text frame

      -  %x2 denotes a binary frame

      -  %x3-7 are reserved for further non-control frames

      -  %x8 denotes a connection close

      -  %x9 denotes a ping

      -  %xA denotes a pong

      -  %xB-F are reserved for further control frames'''
        pass

    @WebsocketFrameProperty
    def MASK(self):
        ''' Mask:  1 bit

      Defines whether the "Payload data" is masked.  If set to 1, a
      masking key is present in masking-key, and this is used to unmask
      the "Payload data" as per Section 5.3.  All frames sent from
      client to server have this bit set to 1.'''
        pass

    @WebsocketFrameProperty
    def PAYLOAD_LENGTH(self):
        '''   Payload length:  7 bits, 7+16 bits, or 7+64 bits

      The length of the "Payload data", in bytes: if 0-125, that is the
      payload length.  If 126, the following 2 bytes interpreted as a
      16-bit unsigned integer are the payload length.  If 127, the
      following 8 bytes interpreted as a 64-bit unsigned integer (the
      most significant bit MUST be 0) are the payload length.  Multibyte
      length quantities are expressed in network byte order.  Note that
      in all cases, the minimal number of bytes MUST be used to encode
      the length, for example, the length of a 124-byte-long string
      can't be encoded as the sequence 126, 0, 124.  The payload length
      is the length of the "Extension data" + the length of the
      "Application data".  The length of the "Extension data" may be
      zero, in which case the payload length is the length of the
      "Application data".'''
        pass

    @WebsocketFrameProperty
    def MASKEY(self):
        '''  Masking-key:  0 or 4 bytes

      All frames sent from the client to the server are masked by a
      32-bit value that is contained within the frame.  This field is
      present if the mask bit is set to 1 and is absent if the mask bit
      is set to 0.  See Section 5.3 for further information on client-
      to-server masking.'''
        pass

    @WebsocketFrameProperty
    def PAYLOAD(self):
        ''' Payload data:  (x+y) bytes

      The "Payload data" is defined as "Extension data" concatenated
      with "Application data".'''
        pass

class WebsocketSession(Session):
    '''
        # Websocket session

        - request          :      BaseHandler request
        - handshake        :       Performs the handshake,--MUST-- be done before any further operation
        - serve            :       Starts serving the client and blocks the thread
            - Performing `run` will block the current request thread until either the server / client decides to close the connection
        - send            :       put message into queue,then it will be sent later if possible
        - receive         :       immediately recieve a frame
        - callback        :       callback for received frame
            - Called once the packet is received
        - shutdown
            - This will set the kill switch

        eg:

            class WebsocketApp(Websocket):
                def callback(self,frame : WebsocketFrame):
                    print('Message:',frame.PAYLOAD)
            @server.rotue(PathMakerModlues.Absoulte('/'))
            def websocket(request):
                ws = WebsocketApp(request)
                ws.handshake()
                ws.serve()
            ...

        The requests,once being adapted into `Websocket` objects,will be added to the `server`'s 
        `websockets` list
    '''
    # Websocket OpCodes
    CONTINUATION = 0x0
    TEXT = 0x1
    BINARY = 0x2
    CLOSE_CONN = 0x8
    PING = 0x9
    PONG = 0xA

    def __init__(self, request,raw_frames=False,*a,**k):
        '''Creates the websocket object

        - `raw_frames`      :       If `true`.the `onRecevie` callback will provide `WebsocketFrames` object instead of `bytearrays`
        (which is the PAYLOAD of the Frame).Note that disabling `raw_frames` will also disable the 
        concatnation of `FIN=0` packets.
        '''        
        super().__init__(request, *a,**k)
        self.raw_frames = raw_frames
        self.keep_alive, self.did_handshake = True, False
        self.__buffer = bytearray()
        # Adds ourself into the server list
        # If the list is not present in the server,create it otherwise
        if not hasattr(self.request.server, 'websockets'):
            setattr(self.request.server, 'websockets', [])
        self.request.raw_request.settimeout(None)
        # no timeouts
        self.request.server.websockets.append(self)
        
    def onNotFound(self):
        '''Useless for WS connections'''
        pass

    def onOpen(self):
        '''Decides what to do once the handshake was perfromed'''
        pass

    def handshake(self):
        '''Do Websocket handshake,must be done first after the request is parsed'''
        self.request.send_response(HTTPStatus.SWITCHING_PROTOCOLS)
        self.request.send_header('Connection', 'Upgrade')
        self.request.send_header(
            'Sec-WebSocket-Accept', 
            self.__ws_gen_responsekey(self.request.headers.get('Sec-WebSocket-Key'))
        )
        self.request.send_header('Upgrade', 'websocket')
        self.request.end_headers()
        self.did_handshake = True
        self.request.log_request('New Websocket session from %s:%s' % self.request.client_address)
        self.onOpen()

    def onReceive(self, frame: Union[bytearray,WebsocketFrame]):
        '''Decides what to do once a frame has been recevied'''
        pass

    def _onReceive(self, frame: WebsocketFrame):
        '''Base recevie method,deals with interal websocket codes and such
        DO NOT OVERRIDE unless you know what you're doing
        '''
        if frame.OPCODE == WebsocketSession.PING:
            # ping -> pong -> end
            self.send(WebsocketFrame(OPCODE=WebsocketSession.PONG))
            return False
        if frame.OPCODE == WebsocketSession.PONG:
            # pong -> ignore -> end
            return False
        elif frame.OPCODE == WebsocketSession.CLOSE_CONN:
            # client requested to close connection -> end
            self.send(WebsocketFrame(OPCODE=WebsocketSession.CLOSE_CONN))
            # accepts such request,raises an exception
            return WebsocketConnectionClosedException(True)
        # All checks passed,proceed to perfrom the callback
        if self.raw_frames:
            return self.onReceive(frame)
        # Return raw frames (including websockets masks) if said so
        # Otherwise,parse the frame and return a bytearray as result
        if not frame.FIN:
            # A Single websocket frame can contain data up to 2^63 Bytes (8192 PB,the MSB's 0)
            # But some clients use FIN bitmask to send little packets over time
            # ...for safety measures.Which means more work needed to be done
            return self.__buffer.extend(frame.PAYLOAD)
            # Extend the buffer
        elif frame.FIN:
            # We have received,the buffer may be cleared next time
            self.__buffer.extend(frame.PAYLOAD)
            self.onReceive(self.__buffer)
            # Perform the real callback
            return self.__buffer.clear()
            # Clear the buffer,Done

    def send(self, frame: WebsocketFrame):
        '''
            Appends message to the buffer,Every message will be constructed as a Websocket Frame
        '''
        return self.request.wfile.write(self.__websocket_constructframe(frame))

    def receive(self) -> WebsocketFrame:
        '''
            Receives a single frame,will block thread
        '''
        return self.__websocket_recieveframe(self.request.rfile)

    def serve(self):
        '''
            Starts processing I/O,and blocks until connection is closed or flag is set


            `ping_interval`   :   WS `Ping` heartbeat interval in seconds,set `0` to disable it
        '''
        if not self.did_handshake:
            raise Exception("Handshake was not performed!")
        
        while self.keep_alive:
            try:                                
                frame = self.receive()
                if not frame:
                    raise WebsocketConnectionClosedException(False)                    
                result = self._onReceive(frame)
                if isinstance(result,WebsocketConnectionClosedException):
                    if result.is_requested:
                        self.keep_alive = False # ends the connection
                    else:raise result           # sends to the error log
                # Checks & runs periodic tasks                
            except Exception as e:
                # Quit once any exception occured
                self.request.log_error(str(e))
                self.keep_alive = False
        self._onClose()

    def close(self):
        '''Tries to close the connection via sending CLOSE_CONN Opcode'''
        self.send(WebsocketFrame(OPCODE=WebsocketSession.CLOSE_CONN))

    def _onClose(self):
        '''Decides what to do once the connection is closed,either by the server or the client'''
        self.set_session() # set it afterwards
        self.request.log_debug('Websocket Connection closed')
        self.request.server.websockets.remove(self)
        # kicks ourself out
        self.onClose()

    def onClose(self):
        pass

    def __websocket_constructframe(self, data: WebsocketFrame) -> bytearray:
        '''
        Constructing frame

        5.2.  Base Framing Protocol:https://tools.ietf.org/html/rfc6455#section-5.2

            0                   1                   2                   3
            0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +-+-+-+-+-------+-+-------------+-------------------------------+
            |F|R|R|R| opcode|M| Payload len |    Extended payload length    |
            |I|S|S|S|  (4)  |A|     (7)     |             (16/64)           |
            |N|V|V|V|       |S|             |   (if payload len==126/127)   |
            | |1|2|3|       |K|             |                               |
            +-+-+-+-+-------+-+-------------+ - - - - - - - - - - - - - - - +
            |     Extended payload length continued, if payload len == 127  |
            + - - - - - - - - - - - - - - - +-------------------------------+
            |                               |Masking-key, if MASK set to 1  |
            +-------------------------------+-------------------------------+
            | Masking-key (continued)       |          Payload Data         |
            +-------------------------------- - - - - - - - - - - - - - - - +
            :                     Payload Data continued ...                :
            + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +
            |                     Payload Data continued ...                |
            +---------------------------------------------------------------+
        '''
        binary = bytearray()
        if not isinstance(data, WebsocketFrame):
            data = WebsocketFrame(PAYLOAD=data)
        binary.append(self.__construct_byte(
            [data.FIN, data.RSV1, data.RSV2, data.RSV3] + self.__extract_byte(data.OPCODE)[4:]))
        # 1st byte:FIN,RSV1,RSV2,RSV3,OPCODE
        if data.PAYLOAD_LENGTH >= 126 and data.PAYLOAD_LENGTH < 65536:
            binary.append(self.__construct_byte(
                [data.MASK] + self.__extract_byte(126)[1:]))
            '''
            If 126, the following 2 bytes interpreted as a 16-bit unsigned integer are the payload length.
            '''
            binary.extend(struct.pack('>H', data.PAYLOAD_LENGTH))
        elif data.PAYLOAD_LENGTH >= 65536 and data.PAYLOAD_LENGTH < 2**63:  # mfw the MSB is always 0
            binary.append(self.__construct_byte(
                [data.MASK] + self.__extract_byte(127)[1:]))
            '''
            If 127, the following 8 bytes interpreted as a 64-bit unsigned integer
            '''
            binary.extend(struct.pack('>Q', data.PAYLOAD_LENGTH))
        elif data.PAYLOAD_LENGTH >= 0 and data.PAYLOAD_LENGTH < 126:
            '''
            Otherwise,the following 7 bytes (after MASK) are interpeted as a 7 bit integer
            '''
            binary.append(self.__construct_byte(
                [data.MASK] + self.__extract_byte(data.PAYLOAD_LENGTH)[1:]))
        else:
            raise Exception("Bad payload length!")
        if data.MASK:
            # Reserved:A server must not mask any frames that it sends to the client
            mkey = self.__gen_maskey()
            data.PAYLOAD = self.__mask(data.PAYLOAD, mkey)
            binary.extend(mkey)
        # 2nd+ bytes: PAYLOAD_LENGTH,MASK,MASKEY
        binary.extend(data.PAYLOAD)
        return binary

    def __websocket_recieveframe(self, rfile: typing.BinaryIO) -> WebsocketFrame:
        '''
        Receiving frame,PAYLOAD is unmasked

        5.2.  Base Framing Protocol:https://tools.ietf.org/html/rfc6455#section-5.2

            0                   1                   2                   3
            0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +-+-+-+-+-------+-+-------------+-------------------------------+
            |F|R|R|R| opcode|M| Payload len |    Extended payload length    |
            |I|S|S|S|  (4)  |A|     (7)     |             (16/64)           |
            |N|V|V|V|       |S|             |   (if payload len==126/127)   |
            | |1|2|3|       |K|             |                               |
            +-+-+-+-+-------+-+-------------+ - - - - - - - - - - - - - - - +
            |     Extended payload length continued, if payload len == 127  |
            + - - - - - - - - - - - - - - - +-------------------------------+
            |                               |Masking-key, if MASK set to 1  |
            +-------------------------------+-------------------------------+
            | Masking-key (continued)       |          Payload Data         |
            +-------------------------------- - - - - - - - - - - - - - - - +
            :                     Payload Data continued ...                :
            + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +
            |                     Payload Data continued ...                |
            +---------------------------------------------------------------+
        '''
        b12 = rfile.read(2)
        if not b12 : return # this would be empty if the socket was closed unexpectedly
        b1, b2 = b12
        FIN, RSV1, RSV2, RSV3 = self.__extract_byte(b1)[:4]
        OPCODE = self.__construct_byte(self.__extract_byte(b1)[4:])
        MASK = self.__get_bit_at(b1, 0)
        PAYLOAD_LENGTH = self.__construct_byte(
            [0] + self.__extract_byte(b2)[1:])
        if PAYLOAD_LENGTH == 126:
            '''
            If 126, the following 2 bytes interpreted as a 16-bit unsigned integer are the payload length.
            '''
            PAYLOAD_LENGTH = struct.unpack('>H', rfile.read(2))[0]
        elif PAYLOAD_LENGTH == 127:
            '''
            If 127, the following 8 bytes interpreted as a 64-bit unsigned integer
            '''
            PAYLOAD_LENGTH = struct.unpack('>Q', rfile.read(8))[0]
        MASKEY = rfile.read(4)
        PAYLOAD = rfile.read(PAYLOAD_LENGTH)
        PAYLOAD = self.__mask(PAYLOAD, MASKEY)
        return WebsocketFrame(FIN, RSV1, RSV2, RSV3, OPCODE, MASK, PAYLOAD_LENGTH, MASKEY, PAYLOAD)

    def __gen_maskey(self):
        '''
            Generate a 32-bit random key
        '''
        return bytearray([random.randint(0, 255) for i in range(0, 4)])

    def __mask(self, d, k):
        '''
        5.3.  Client-to-Server Masking:https://tools.ietf.org/html/rfc6455#section-5.3

        Octet i of the transformed data ("transformed-octet-i") is the XOR of

        octet i of the original data ("original-octet-i") with octet at index

        i modulo 4 of the masking key ("masking-key-octet-j")

                j                           =                          i MOD 4
                transformed-octet-i = original-octet-i XOR masking-key-octet-j

        To get DECODED, loop through the octets (bytes a.k.a. characters for text data) of ENCODED and XOR the octet with the (i modulo 4)th octet of MASK.
        '''
        return bytearray([d[i] ^ k[i % 4] for i in range(0, len(d))])

    def __construct_byte(self, a):
        '''
            Form an array of 1s and 0s,gets a byte (little-endian)
        '''
        return int(''.join([str(i) for i in a]), 2)

    def __extract_byte(self, a):
        '''
            Turns a byte into an array of 1s and 0s            
        '''
        return [self.__get_bit_at(a, i) for i in range(0, 8)]

    def __get_bit_at(self, a, i):
        '''
            Get bit at certain poistion,stars from the highest bit indexed 0
        '''
        return ((a << i) & 128) >> 7

    def __ws_gen_responsekey(self, key):
        '''
        As described in RFC6455:https://tools.ietf.org/html/rfc6455#section-1.3

            Sec-WebSocket-Key: dGhlIHNhbXBsZSBub25jZQ==

        For this header field, the server has to take the value (as present

        in the header field, e.g., the base64-encoded [RFC4648] version minus

        any leading and trailing whitespace) and concatenate this with the

        Globally Unique Identifier (GUID, [RFC4122]) "258EAFA5-E914-47DA-

        95CA-C5AB0DC85B11" in string form, which is unlikely to be used by

        network endpoints that do not understand the WebSocket Protocol.  A

        SHA-1 hash (160 bits) [FIPS.180-3], base64-encoded (see Section 4 of

        [RFC4648]), of this concatenation is then returned in the server's

        handshake.
        '''
        GUID = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
        rkey = key + GUID    # contact in string form
        # returns SHA-1 hash and b64 encoded concatenation
        rkey = base64.b64encode(hashlib.sha1(rkey.encode()).digest())
        return rkey.decode()


@ModuleWrapper
def WebsocketSessionWrapper(raw_frames=False,**kwargs):
    '''Wrapper for websocket requests

    Usage:

        ...
        from pywebhost import Websocket
        class WSApp(Websocket):
            def onOpen(self):
                ...
            def onClose(self):
                ...
            def onRecevie(self):
                ...
        @server.route('/ws')
        @WebsocketSessionWrapper(raw_frames=False)
        def wsapp(request : Request,content):
            return WSApp
    Args:

        raw_frames (bool, optional): To receive concatnated content as bytearrary or raw websocket frames. Defaults to False.
    '''
    def suffix(request,function_result : WebsocketSession):
        if not function_result:
            return # Dropping connection
        if not issubclass(function_result,WebsocketSession):
            raise TypeError('A `pywebhost.WebsocketSession` (sub) class is required') 
        # Making the websocket connection
        session = function_result(request,raw_frames=raw_frames,**kwargs)
        session.handshake()
        session.serve()
    return None , suffix


from http.client import SERVICE_UNAVAILABLE
from pywebhost.modules import BadRequestException
import selectors,socketserver,sys

from .handler import Request
from .modules import *
# from .modules import *
from re import fullmatch
from http import HTTPStatus

__version__ = '1.0.7'

class PathMaker(dict):
    '''For storing and handling path mapping
    
        The keys and values are stored as regex pattern strings
        Keys are used to check is the target URL matching the stored URL,which,using regexes will be a great idea

        To set an item:

            pathmaker['/.*'] = lambda a:SendFile('index.html')

        The server will be finding the functions simply with this:

            pathmaker['/']()

    '''
    def __init__(self):
        super().__init__()

    def __setitem__(self, pattern, value):
        '''Sets an path to be routed'''
        if not isinstance(pattern,str):raise Exception('The keys & values must be regexes string')
        super().__setitem__(pattern,value)

    def __getitem__(self, key):
        '''Iterates all keys to find matching one

        The last one added has a better piority of getting called
        '''
        for pattern in list(self.keys())[::-1]: # LIFO
            if fullmatch(pattern,key):
                yield super().__getitem__(pattern)

class PyWebHost(socketserver.ThreadingMixIn, socketserver.TCPServer,):
    '''
        # PyWebHost
        
        To start a server:

            server = PyWebHost(('',1234))
            server.serve_forever()

        You can test by typing `http://localhost:1234` into your browser to retrive a glorious error page ((
    '''
    daemon_threads = True
    def handle_error(self, request : Request, client_address):
        """Handle an error gracefully.  May be overridden.

        By default,it prints the latest stack trace
        """
        super().handle_error(request,client_address)


    def serve_forever(self, poll_interval=0.5):
            """Handle one request at a time until shutdown.

            Polls for shutdown every poll_interval seconds. Ignores
            self.timeout. If you need to do periodic tasks, do them in
            another thread.
            """
            self._BaseServer__is_shut_down.clear()
            try:
                # XXX: Consider using another file descriptor or connecting to the
                # socket to wake this up instead of polling. Polling reduces our
                # responsiveness to a shutdown request and wastes cpu at all other
                # times.
                with selectors.SelectSelector() as selector:
                    selector.register(self, selectors.EVENT_READ)
                    while not self._BaseServer__shutdown_request:
                        ready = selector.select(poll_interval)                        
                        # bpo-35017: shutdown() called during select(), exit immediately.
                        if self._BaseServer__shutdown_request:
                            break
                        if ready:
                            self._handle_request_noblock()
                        self.service_actions()
            finally:
                self._BaseServer__is_shut_down.set()

    def __handle__(self, request : Request):
        '''
        Maps the request with the `PathMaker`
        
        The `request` is provided to the router
        '''
        for method in self.paths[request.path]:
            try:
                return method(request,None)
                # Succeed,end this handle call
            except BadRequestException as e:
                # For Other server-side exceptions,let the client know
                return request.send_error(e.code,e.explain)
            except Exception as e:
                return request.send_error(SERVICE_UNAVAILABLE,explain='There was an error processing your request:%s'%e)
        # Request's not handled:No URI matched
        return request.send_error(HTTPStatus.NOT_FOUND)

    def route(self,pattern):
        '''
        Routes a HTTP Request

        e.g:

            @server.route('/')
                def index():lambda a:SendFile('index.html')
        '''
        def wrapper(method):
            self.paths[pattern] = method
            return method
        return wrapper

    def format_error_message(self,code:int,message:str,explain:str,request:Request):
        return f'''
        <head>        
            <title>PyWebHost Error - {self.protocol_version} {code}</title>
        ''' + '''<style>i {position : fixed;bottom:0%;left : 0%;font-size: 14px;}</style>''' + f'''</head><body>
        <div>
            <center><h1>{code} {message}</h1></center>
            <hr><center>{explain}</center><hr>
        </div>
        <i>PyWebHost {__version__}  on {sys.version}</i>
        </body>
        '''

    def __init__(self, server_address : tuple):
        self.paths = PathMaker()
        # A paths dictionary which has `lambda` objects as keys
        self.protocol_version = "HTTP/1.1"
        # What protocol version to use.
        # Here's a note:
        # HTTP 1.1 will automaticly keep the connections alive,so
        # `close_connection = True` needs to be called once the connection is done

        # Error page format. %(`code`)d %(`message`)s %(`explain`)s are usable
        super().__init__(server_address, Request)



from random import randrange
from ..handler import Request
from . import ModuleWrapper, WriteContentToRequest, writestream
import time,random

from hashlib import md5
SESSION_KEY = 'sess_'

_sessions = {}
class Session(dict):

    @staticmethod
    def mapUri(url,object_):
        '''From the request path to local method'''
        classpath = url.replace('/','_')
        if hasattr(object_,classpath):
            return getattr(object_,classpath)

    @property
    def new_uid(self):
        '''Generates new uid'''
        str_ = '%sDamn%sI%sOughtTorethinkthis' % (SESSION_KEY,time.time() * 1e6,randrange(1e7,1e8-1))
        return md5(str_.encode()).hexdigest()

    @property
    def session_id(self):
        '''The UID of the session

        Returns:
            str : UID
        '''
        if not self.use_session_id:return None
        session_id = self.request.cookies.get(SESSION_KEY) or self.request.cookies_buffer.get(SESSION_KEY)
        if not session_id:
            session_id = self.new_uid
            self.request.send_cookies(SESSION_KEY,session_id)
            return session_id
        return session_id.value
    
    def get_session(self):
        '''Gets session dictionary by uid,maybe overridden
        
        Returns:
            dict : The such dict
        '''
        if self.session_id:
            if not self.session_id in _sessions.keys():_sessions[self.session_id] = {}
            return _sessions[self.session_id]
        return {}
    def set_session(self):
        '''Saves session dictionary by updating it with our values'''
        if self.session_id:
            if not self.session_id in _sessions.keys():_sessions[self.session_id] = {}
            _sessions[self.session_id].update(self)

    def onError(self,error):
        '''What to do when an execption occured

        Args:
            error (Execption): The said exception
        '''
        pass

    def onNotFound(self):
        '''What to do when the path cannot be mapped'''
        self.request.send_error(404)

    def onOpen(self):
        '''What to do when the session starts'''
        self.request.send_response(200)

    def onClose(self):
        '''What to do when the session ends'''
        self.request.flush_headers()        

    def __init__(self,request : Request,use_session_id=True) -> None:           
        super().__init__()
        self.request = request    
        self.use_session_id = use_session_id
        # try to map the request path to our local path
        self.request_func = Session.mapUri(self.request.path,self)
        self.update(self.get_session()) # loads session dict
        try:            
            if not self.request_func or not '_' in self.request_func.__name__:
                self.onNotFound()
            else:                
                self.onOpen()
                self.request_func_result = self.request_func()                
                self.set_session()            # saves session dict
                self.onClose()
            # calls the request func                
        except Exception as e:
            self.onError(e)        
        
@ModuleWrapper
def SessionWrapper():
    def suffix(request,function_result):        
        if not issubclass(function_result,Session):
            raise TypeError("Request did not return a Session (or its subclass) object.")
        return function_result(request)
        # Enters the session
    return None,suffix


from http.client import OK, ResponseNotReady
from http.cookies import SimpleCookie
import logging,socket
from datetime import datetime
from socketserver import StreamRequestHandler
from http import HTTPStatus,client,cookies
from html import escape
from urllib.parse import urlparse,parse_qs,unquote
from io import BufferedIOBase, IOBase
import time
'''
Essentially static class variables
'''

# The default request version.  This only affects responses up until
# the point where the request line is parsed, so it mainly decides what
# the client gets back when sending a malformed request line.
# Most web servers default to HTTP 0.9, i.e. don't send a status line.
default_request_version = "HTTP/0.9"
default_response_code = OK

_MAXLINE = 65536
_MAXHEADERS = 100
_MAXTIMEOUT = 60

class Headers(dict):

    def encode(self):
        return str(self).encode()

    def __str__(self) -> str:
        # formatting the lines
        buffer = []        
        if self.response_line:buffer.append(self.response_line )
        for k,v in self.items():buffer.append('%s: %s' % (k,v))
        str_ = buffer[0] + '\n'.join(buffer[1:]) + '\r\n\r\n'
        
        return str_

    def __init__(self,response_line='') -> None:
        self.response_line = response_line

    def add_header_line(self,header_line : bytes):
        if not isinstance(header_line,str):
            header_line = header_line.decode()
        header_line = header_line.strip().split(':',maxsplit=1)
        # trying to decode the headers
        if len(header_line) < 2 : return
        key,value = header_line
        self[key.strip()] = value.strip()
        return key,value

    @staticmethod
    def parse(fp : IOBase):
        """Parses only RFC2822 headers from a file pointer.
        """
        headers = Headers()
        while True:
            line = fp.readline(_MAXLINE + 1)
            if len(line) > _MAXLINE:
                raise client.LineTooLong("header line")
            headers.add_header_line(line)
            if len(headers) > _MAXHEADERS:
                raise client.HTTPException("got more than %d headers" % _MAXHEADERS)
            if line in (b'\r\n', b'\n', b''):
                break      
        return headers

class Request(StreamRequestHandler):
    '''`http` module's handler with extra perks & optimization'''
    """
    Base properties
    """

    wfile : BufferedIOBase
    '''`BufferedIOBase` like I/O for writing messages to sockets'''
    rfile : BufferedIOBase
    '''`BufferedIOBase` like I/O for reading messages from sockets'''
    headers : Headers
    '''Contains parsed headers'''        
    headers_buffer : Headers
    '''The headers to be parsed'''
    cookies : cookies.SimpleCookie
    '''Contains request cookies'''
    cookies_buffer : cookies.SimpleCookie
    '''The cookies to be sent by us'''
    command : str
    '''The request command (GET,POST,etc)'''
    raw_requestline : str
    '''Raw `HTTP` request line of the request'''
    raw_request : socket.socket
    '''Raw TCP socket'''
    close_connection : bool
    '''Whether to preserve the connection (keep-alive one) or not'''
    def __init__(self, request, client_address, server):
        '''The `server`,which is what instantlizes this handler,must have `__handle__` method
        which takes 1 argument (for the handler itself) 
        '''
        self.logger = logging.getLogger('Request')
        '''These values are from the server'''
        # The version of the HTTP protocol we support.
        # Set this to HTTP/1.1 to enable automatic keepalive
        self.protocol_version = server.protocol_version
        # Error page formats
        self.format_error_message = server.format_error_message
        # hack to maintain backwards compatibility
        self.responses = {
            v: (v.phrase, v.description)
            for v in HTTPStatus.__members__.values()
        }
        self.headers = Headers()
        self.headers_buffer = Headers()
        self.cookies = SimpleCookie()
        self.cookies_buffer = SimpleCookie()
        self.raw_request = request
        self.raw_request.settimeout(_MAXTIMEOUT)
        super().__init__(request, client_address, server)

    def parse_request(self):
        """Parse a request (internal).

        The request should be stored in self.raw_requestline; the results
        are in self.command, self.path, self.request_version and
        self.headers.

        Return True for success, False for failure; on failure, any relevant
        error response has already been sent back.
        """
        self.command = None  # set in case of error on the first line
        self.request_version = default_request_version
        self.close_connection = True
        try:
            requestline = self.raw_requestline.decode()    
        except UnicodeDecodeError as e:
            # Bad request,this is mostly caused by bad SSL support
            self.log_error('Bad request line: %s',e)
            self.close_connection = True
            return False
        requestline = requestline.rstrip('\r\n')
        self.requestline = requestline
        words = requestline.split()
        if len(words) == 0:  # Empty request line?
            return False
        if len(words) >= 3:  # Enough to determine protocol version
            version = words[-1]
            try:
                if not version.startswith('HTTP/'):
                    raise ValueError
                self.base_version_number = version.split('/', 1)[1]
                version_number = self.base_version_number.split(".")
                # RFC 2145 section 3.1 says there can be only one "." and
                #   - major and minor numbers MUST be treated as
                #      separate integers;
                #   - HTTP/2.4 is a lower version than HTTP/2.13, which in
                #      turn is lower than HTTP/12.3;
                #   - Leading zeros MUST be ignored by recipients.
                if len(version_number) != 2:
                    raise ValueError
                version_number = int(version_number[0]), int(version_number[1])
            except (ValueError, IndexError):
                self.send_error(HTTPStatus.BAD_REQUEST,"Bad request version (%r)" % version)
                return False
            if version_number >= (1, 1) and self.protocol_version >= "HTTP/1.1":
                self.close_connection = False
            if version_number >= (2, 0):
                self.send_error(HTTPStatus.HTTP_VERSION_NOT_SUPPORTED,"Invalid HTTP version (%s)" % self.base_version_number)
                return False
            self.request_version = version

        if not 2 <= len(words) <= 3:
            self.send_error( HTTPStatus.BAD_REQUEST,"Bad request syntax (%r)" % requestline)
            return False
        command, path = words[:2]
        if len(words) == 2:
            self.close_connection = True
            if command != 'GET':
                self.send_error(HTTPStatus.BAD_REQUEST,"Bad HTTP/0.9 request type (%r)" % command)
                return False        
        self.command, self.raw_path = command, path
        self.scheme, self.netloc, self.path, self.params, self.query, self.fragment = urlparse(self.raw_path)
        self.path = unquote(self.path)
        self.query = parse_qs(self.query) # Decodes query string to a `dict`
        # Decode the URI
        # Examine the headers and look for a Connection directive.
        try:
            self.headers = Headers.parse(self.rfile)
            if self.headers.get('Cookie'):
                cookies = self.headers.get('Cookie').replace(' ','%20') # esacpe spaces : https://tools.ietf.org/html/rfc6265#section-4.1.1
                self.cookies = SimpleCookie(cookies)                
        except client.LineTooLong as err:
            self.send_error(
                HTTPStatus.REQUEST_HEADER_FIELDS_TOO_LARGE,
                "Line too long",
                str(err))
            return False
        except client.HTTPException as err:
            self.send_error(
                HTTPStatus.REQUEST_HEADER_FIELDS_TOO_LARGE,
                "Too many headers",
                str(err)
            )
            return False                    
        # Decode the headers,cookies         
        conntype = self.headers.get('Connection')
        if conntype:
            if conntype.lower() == 'close':
                self.close_connection = True
            elif (conntype.lower() == 'keep-alive' and
                self.protocol_version >= "HTTP/1.1"):
                self.close_connection = False
        # Examine the headers and look for an Expect directive
        expect = self.headers.get('Expect')
        if (expect and expect.lower() == "100-continue" and
                self.protocol_version >= "HTTP/1.1" and
                self.request_version >= "HTTP/1.1"):
            if not self.handle_expect_100():
                return False
        return True

    def handle_expect_100(self):
        """Decide what to do with an "Expect: 100-continue" header.

        If the client is expecting a 100 Continue response, we must
        respond with either a 100 Continue or a final response before
        waiting for the request body. The default is to always respond
        with a 100 Continue. You can behave differently (for example,
        reject unauthorized requests) by overriding this method.

        This method should either return True (possibly after sending
        a 100 Continue response) or send an error response and return
        False.
        """
        self.send_response_only(HTTPStatus.CONTINUE)
        self.end_headers()
        return True

    def handle_one_request(self):
        """Handle a single HTTP request.

        You normally don't need to override this method; see the class
        __doc__ string for information on how to handle specific HTTP
        commands such as GET and POST.
        """
        try:
            self.raw_requestline = self.rfile.readline(65537)
            if len(self.raw_requestline) > 65536:
                self.requestline = ''
                self.request_version = ''
                self.command = ''
                self.send_error(HTTPStatus.REQUEST_URI_TOO_LONG)
                return
            if not self.raw_requestline:
                self.close_connection = True
                return
            if not self.parse_request():
                # An error code has been sent, just exit
                return
            '''Now,ask the server to process the request'''
            self.server.__handle__(request=self)
            if self.headers_buffer.response_line or self.headers_buffer:
                raise ResponseNotReady('Response header lines were not flushed')        
            return
        except ResponseNotReady as e:
            self.log_error("Bad Response: %s",e)
            self.send_error(HTTPStatus.SERVICE_UNAVAILABLE)
            self.close_connection = True
            return
        except socket.timeout as e:
            #a read or a write timed out.  Discard this connection
            self.log_error("Request timed out: %r", e)
            self.close_connection = True
            return

    def handle(self):
        """Handle multiple requests if necessary."""
        self.close_connection = True
        self.handle_one_request()            
        while not self.close_connection:           
            self.handle_one_request()          
    def send_error(self, code, message=None, explain=None):
        """Send and log an error reply.

        Arguments are
        * code:    an HTTP error code
                   3 digits
        * message: a simple optional 1 line reason phrase.
                   *( HTAB / SP / VCHAR / %x80-FF )
                   defaults to short entry matching the response code
        * explain: a detailed message defaults to the long entry
                   matching the response code.

        This sends an error response (so it must be called before any
        output has been generated), logs the error, and finally sends
        a piece of HTML explaining the error to the user.
        """

        try:
            shortmsg, longmsg = self.responses[code]
        except KeyError:
            shortmsg, longmsg = '???', '???'
        if message is None:
            message = shortmsg
        if explain is None:
            explain = longmsg
        self.log_error("HTTP %d -- %s", code, explain)
        self.clear_header()
        self.send_response_only(code, message)
        self.send_header('Connection', 'close')

        # Message body is omitted for cases described in:
        #  - RFC7230: 3.3. 1xx, 204(No Content), 304(Not Modified)
        #  - RFC7231: 6.3.6. 205(Reset Content)
        body = None
        if (code >= 200 and
            code not in (HTTPStatus.NO_CONTENT,
                         HTTPStatus.RESET_CONTENT,
                         HTTPStatus.NOT_MODIFIED)):
            # HTML encode to prevent Cross Site Scripting attacks
            # (see bug #1100201)
            content = self.format_error_message(
                code   =code,
                message=escape(message, quote=False),
                explain=escape(explain, quote=False),
                request=self
            )
            body = content.encode('UTF-8', 'replace')
            self.send_header("Content-Type", 'text/html;charset=utf-8')
            self.send_header('Content-Length', str(len(body)))
        self.end_headers()

        if self.command != 'HEAD' and body:
            self.wfile.write(body)
    
    def response_line(self,code,message=''):
        return "%s %d %s\r\n" % (self.protocol_version, code, message)

    def send_response(self, code, message=None):
        """Add the response header to the headers buffer and log the
        response code.
        """
        self.log_request(code)
        self.send_response_only(code, message)

    def send_response_only(self, code, message=None):
        """Send the response header only."""
        if self.request_version != 'HTTP/0.9':
            if message is None:
                if code in self.responses:
                    message = self.responses[code][0]
                else:
                    message = ''
            self.headers_buffer.response_line = self.response_line(code,message)
            # Always send this at the begining
    
    def clear_header(self):
        self.headers_buffer = Headers()

    def send_header(self, keyword, value):
        """Send a MIME header to the headers buffer."""
        if self.request_version != 'HTTP/0.9':
            self.headers_buffer[keyword] = value

        if keyword.lower() == 'connection':
            if value.lower() == 'close':
                self.close_connection = True
            elif value.lower() == 'keep-alive':
                self.close_connection = False

    def send_cookies(self,key,value):
        '''Sets a cookie'''
        self.cookies_buffer[key]=value

    def end_headers(self):
        """Adds the cookies and blank line ending of the MIME headers to the buffer,
        then flushes the buffer"""
        if self.request_version != 'HTTP/0.9':         
            if self.cookies_buffer:self.headers_buffer.add_header_line(self.cookies_buffer.output())
            self.flush_headers()

    def flush_headers(self):
        if not self.headers_buffer.response_line:
            raise Exception('Request was not responed with `requst.send_response`.')    
        self.wfile.write(self.headers_buffer.encode())
        self.headers_buffer = Headers()

    def log_request(self, code='-'):
        """Log an accepted request.

        This is called by send_response()
        """
        if isinstance(code, HTTPStatus):
            code = code.value
        self.log_debug('"%s" %s',self.requestline, str(code))

    def address_string(self):
        """Return the client address."""
        return self.client_address[0] + ':' + str(self.client_address[1])

    def useragent_string(self):
        """Returns the client UA header,if applicable"""
        return self.headers.get('User-Agent')

    def time_string(self):
        return str(datetime.now())

    def format_log(self,format,*args):
        """
        Formats a logging message

        Takes `format` and `args` which will construct the base message
        ...and adds other componet into the string

        This method CAN be overwritten.This tries to mimic the
        NGINX Style logging,which looks like this:

            {Client Address} [{Time}] "{Verb} {Path} {HTTP Version}" {Message} {User-Agent}
        """
        try:
            return f'{self.address_string()} [{self.time_string()}]' + f'"{self.command} {self.path} {self.base_version_number}" {format % args} {self.useragent_string()}'
        except Exception:
            # If an exception ouccred,fallback to another format            
            return f'{self.address_string()} [{self.time_string()}] {format % args}'
    
    def log_error(self, format, *args):
        """Log an error.

        This is called when a request cannot be fulfilled.  By
        default it passes the message on to log_message().

        Arguments are the same as for log_message().
        """
        self.logger.warning(self.format_log(format,*args))        

    def log_debug(self, format, *args):
        """Log a debug message.

        This is called when a request is done

        Arguments are the same as for log_message().
        """
        self.logger.debug(self.format_log(format,*args))        

    def log_message(self, format, *args):
        """Log an arbitrary message.

        This is used by all other logging functions.  Override
        it if you have specific logging wishes.

        The first argument, FORMAT, is a format string for the
        message to be logged.  If the format string contains
        any % escapes requiring parameters, they should be
        specified as subsequent arguments (it's just like
        printf!).

        The formats are decided by `format_log`
        """
        self.logger.info(self.format_log(format,*args))
    
import sched


from codecs import decode
from http.client import CONTINUE, OK
from io import BufferedIOBase, BufferedReader, IOBase,BytesIO
from os import system
import time
from ..handler import Request
from http import HTTPStatus
import os,mimetypes,json,base64,select
from typing import Any, NamedTuple, Type, Union

class BadRequestException(Exception):
    '''Exception with code and explaination'''
    def __init__(self,code,explain=None) -> None:
        self.code = code
        self.explain = explain
        super().__init__(explain)

def ModuleWrapper(provider):    
    '''Base circlular wrapper support func

    Args:
        provider (function)

    Usage:
    - As a module
    @ModuleWrapper
    def dummy_func(dummy_arg):
        def prefix(request,previous_prefix_result):
            ...
        def suffix(request,function_result):
            ...
        return prefix,suffix
    - As a wrapped module
    @dummy_func(dummy_arg=42)
    def dummy_request(request,content):
        ...

    # provider
     A function,that when called with arguments,always returns tuple (prefix,suffix) of functions:
    - prefix : what to execute when called by server
        - takes two arguments : `request` (Request) , `previous_prefix_result`
            - `request` is what it is,represented by the `pywebhost.Request` object
            - `previous_prefix_result` is the result of the prefix function of the previous wrapper
        - returns a value , which is used by the function wrapped by `RequestFunctionWrapper`
    - suffix : what to do after the `RequestFunctionWrapper` is finished
        - takes two arguments : `request` (Request) , `function_result`
            - `request` is what it is,represented by the `pywebhost.Request` object
            - `function_result` is the result of the function wrapped
        - returns a value , which , if chained with another wrapper , will be its `function_result`
    '''
    def UserWrapper(*a,**k):
        prefix,suffix = provider(*a,**k)
        def RequestFunctionWrapper(function):
            def RequestWrapper(request : Request,previous_prefix_result=None):
                prefix_result   = prefix  (request,previous_prefix_result) if prefix else previous_prefix_result
                function_result = function(request,prefix_result)
                suffix_result   = suffix  (request,function_result) if suffix else function_result
                return suffix_result
            return RequestWrapper
        return RequestFunctionWrapper
    return UserWrapper

def any2bytes(any):
    if isinstance(any,str):return any.encode()
    return bytearray(any)

def any2str(any):
    if isinstance(any,bytearray) or isinstance(any,bytes):return decode(any)
    return str(any)

def filesize(path):
    return os.path.getsize(path)

def streamcopy(from_:IOBase,to_:IOBase,size=-1,chunk_size=163840):
    '''Copies content from one buffer to the other,chunk by chunk
    
    Args:
        from_ (IOBase): Stream to copy from
        to_ (IOBase): Stream to copy to
        size (int, optional): Length to be copied. Defaults to -1.
        chunk_size (int, optional): Size of chunk. Defaults to 163840.

    Returns:
        int : copied length
    '''
    if not size:return 0
    size,copied = int(size),0
    if size < 0:
        # read until EOF
        def copychunk():
            chunk = from_.read(chunk_size)
            if not chunk:return 0
            to_.write(chunk)
            return len(chunk)
        while (True):
            copied_ = copychunk()
            if not copied_:break
            copied+=copied_                
    else:
        # read `size` of bytes
        for offset in range(0,size,chunk_size):
            remaining = size - offset 
            chunk = from_.read(remaining if remaining < chunk_size else chunk_size)
            if not chunk:break
            copied += len(chunk)
            to_.write(chunk)
    return copied

def readstream(request:Request):
    '''Reads all the content from client

    Args:
        request (Request): Request

    Raises:
        BadRequestException: When `Content-Length` is not defined

    Returns:
        bytes : Read content
    '''
    buffer = BytesIO()
    length = request.headers.get('Content-Length')
    if not length:raise BadRequestException(400) # this header musn't be empty : https://tools.ietf.org/html/rfc2616#page-33
    streamcopy(request.rfile,buffer,length)
    buffer.seek(0)
    decoded = buffer.read()
    return decoded

def writestream(request:Request,data):
    '''Writes content to client

    Args:
        request (Request): Request
        data (Any): bytes / str like objects

    Returns:
        int : Bytes written
    '''
    buffer  = any2bytes(data)
    request.send_header('Content-Length',str(len(buffer)))
    request.end_headers()    
    sent    = streamcopy(BytesIO(buffer),request.wfile)
    return sent

def Redirect(request:Request,redirect_to:str,code:int = HTTPStatus.FOUND) -> None:
    '''Redirects request to another URL

    Args:
        redirect_to (str): Where to redirect to
        code (int, optional): MOVED_PERMANENTLY or FOUND. Defaults to HTTPStatus.FOUND.
    '''
    request.send_response(code)
    request.send_header('Location',redirect_to)
    request.end_headers()

def ReadContentToBuffer(request:Request,stream_to:IOBase,chunk_size : int=163840):
    '''Reads content of request to buffer

    Args:
        request (Request): Request
        stream_to (IOBase): Where to write to
        chunk_size (int, optional): Size of chunk. Defaults to 163840.

    Returns:
        int : Read bytes
    '''
    length = request.headers.get('Content-Length')
    assert length # this header musn't be empty : https://tools.ietf.org/html/rfc2616#page-33
    return streamcopy(request.rfile,stream_to,length,chunk_size)

def WriteContentToRequest(
    request : Request,
    object:Union[str,bytes,bytearray],
    partial_acknowledge : bool=False,
    length : int=-1,
    chunk_size : int=163840,
    mime_type : str='text/plain') -> None:
    '''Writes buffer / file as static content to stream

    Args:
        request (Request): Request
        object (Union[str,bytes,bytearray]): Content to be sent
        partial_acknowledge (bool, optional): React to HTTP 206s or not. Defaults to False.
        length (int, optional): The length to be sent. Defaults to -1.
        chunk_size (int, optional): Size of chunk. Defaults to 163840.
        mime_type (str, optional): Content mime type. Defaults to 'text/plain'.

    Returns:
        int : Bytes sent
    '''
    # serializing stream
    if isinstance(object,str):
        # str - file path
        length = filesize(object) if length < 0 else length
        stream = open(object,'rb')
    elif hasattr(object,'read'):
        # readable - IO-like objects
        stream = object
        
    def send_once(request):                        
        request.send_response(HTTPStatus.OK)   
        if length > 0:request.send_header('Content-Length',length)
        request.send_header('Content-Type',mime_type)
        request.end_headers()     
        streamcopy(stream,request.wfile,chunk_size=chunk_size)
        return True

    def send_range(request):
        Range = request.headers.get('Range')
        if not Range:return False # no range header
        if not Range[:6] == 'bytes=' : return request.send_error(HTTPStatus.BAD_REQUEST,'Only ranges of `bytes` are supported')
        start,end = Range[6:].split('-')
        start,end = int(start if start else 0),int(end if end else length)
        if not (start >= 0 and start < length and end > 0 and end > start and end <= length):
            # Range not satisfiable
            return request.send_error(HTTPStatus.REQUESTED_RANGE_NOT_SATISFIABLE)
        
        request.clear_header() # re-constructs headers
        request.send_response(HTTPStatus.PARTIAL_CONTENT)
        request.send_header('Accept-Ranges','bytes')
        request.send_header('Content-Length',str(end - start))
        request.send_header('Content-Type',mime_type)
        request.send_header('Content-Range','bytes %s-%s/%s' % (start,end,length))
        request.end_headers()          

        stream.seek(start)
        streamcopy(stream,request.wfile,end - start,chunk_size=chunk_size)
        return True
    
    if partial_acknowledge:
        if length > 0:
            request.send_header('Accept-Ranges','bytes')
            if send_range(request):return True
    return send_once(request)

@ModuleWrapper
def VerbRestrictionWrapper(verbs : list = ['GET','POST']) -> None:
    '''Wrapper to limit request to a certain range of verbs

    When request's verb is not in the list,a HTTP 400 will be responded instead

    Usage:

        ...
        @server.route('/api/get_v1')
        @VerbRestrictionWrapper(['POST'])
        def get(request : Request,content):
            ...

    Args:
        verbs (list, optional): [description]. Defaults to ['GET','POST'].
    '''
    def prefix(request,previous_suffix):
        '''Restricts HTTP Verbs,does nothing if the verb is in the `verbs` list'''
        if not request.command in verbs:
            raise BadRequestException(HTTPStatus.BAD_REQUEST,'Verb %s is not allowed' % request.command)
    return prefix , None

@ModuleWrapper
def BinaryMessageWrapper(read=True,write=True):
    '''Wrapper to receive,send binary content before,after the request
    
    Usage:
    
        @server.route('.*')        
        @BinaryMessageWrapper(read=True,write=True)
        def main(request:Request,content):
            return salteddigest(content)

    Args:
        read (bool, optional): Read the content. Defaults to True.
        write (bool, optional): Write the content. Defaults to True.
    '''
    def prefix(request,previous_prefix_result):
        if not read:return None
        binary = readstream(request) if previous_prefix_result is None else previous_prefix_result    
        return binary
    def suffix(request,function_result):
        if function_result:
            binary = any2bytes(function_result)
            if write:writestream(request,binary)
            return function_result
        return bytearray()
    return prefix , suffix

@ModuleWrapper
def JSONMessageWrapper(decode=True,encode=True,read=True,write=True):
    '''Wrapper to receive,send JSON content before,after the request

    Usage:
    
        @server.route('.*')        
        @Base64MessageWrapper(decode=True,encode=False,read=True,write=False) # only decodes content
        @JSONMessageWrapper(decode=True,encode=True,write=True,read=False)    # parses decoded content,encodes return value,then writes it
        def main(request:Request,content):
            return {'data':'senstive_data_or_sth_idk'}


    Args:
        decode (bool, optional): To decode the request. Defaults to True.
        encode (bool, optional): To encode the response. Defaults to True.
        read (bool, optional): Read the content. Defaults to True.
        write (bool, optional): Write the content. Defaults to True.
    '''
    def prefix(request,previous_prefix_result):
        if not read:return None
        string = any2str(readstream(request)) if previous_prefix_result is None else previous_prefix_result
        return string if not decode else json.loads(string)
    def suffix(request,function_result):
        if function_result and encode:
            string = json.dumps(function_result)
            if write:writestream(request,string)
            return string
        elif function_result:
            return function_result
        else:
            return None
    return prefix , suffix

@ModuleWrapper
def Base64MessageWrapper(decode=True,encode=True,read=True,write=True):
    '''Wrapper to receive,send BASE64 content before,after the request

    Usage:
    
        @server.route('.*')
        @JSONMessageWrapper(decode=True,encode=False,write=False)     # only reads request as pass it as `content`
        @Base64MessageWrapper(decode=False,read=False)                # encodes the response and writes it
        def main(request:Request,content):
            return 'senstive_data_or_sth_idk'    


    Args:
        decode (bool, optional): To decode the request. Defaults to True.
        encode (bool, optional): To encode the response. Defaults to True.
        read (bool, optional): Read the content. Defaults to True.
        write (bool, optional): Write the content. Defaults to True.
    '''    
    def prefix(request,previous_prefix_result):
        if not read:return None
        binary = any2str(readstream(request)) if previous_prefix_result is None else previous_prefix_result
        return binary if not decode else base64.b64decode(binary)
    def suffix(request,function_result):
        if function_result:
            binary = function_result.encode()
            if encode:base64.b64encode(binary)
            if write:writestream(request,binary)
            return binary
        elif function_result:
            return function_result
        else:
            return None
    return prefix , suffix